"""Robot URDF visualizer

Requires yourdfpy and robot_descriptions. Any URDF supported by yourdfpy should work.
- https://github.com/robot-descriptions/robot_descriptions.py
- https://github.com/clemense/yourdfpy

The :class:`viser.extras.ViserUrdf` is a lightweight interface between yourdfpy
and viser. It can also take a path to a local URDF file as input.
"""

from __future__ import annotations

import json
import logging
import math
import pathlib
import subprocess
import tempfile
import time
import urllib.parse
from functools import partial
from typing import Literal, Optional
import os
import io

import numpy as np
import uvicorn
import yourdfpy

import viser
import requests
from fastapi import FastAPI, File, UploadFile
from fastapi.responses import FileResponse, PlainTextResponse
from viser.extras import ViserUrdf

from . import setup_logging

setup_logging("punch.log", print_console=True)
logger = logging.getLogger(__name__)

app = FastAPI()


# 下载文件
@app.get("/resources/{filename}")
async def download_file(filename: str):
    file_path = f"./resources/{filename}"
    with open(file_path, "r") as file:
        content = file.read()
    return PlainTextResponse(content)


def make_resource_dir(suffix):
    os.makedirs("resources", exist_ok=True)

    _, temp_file = tempfile.mkstemp(suffix=suffix)

    return f"resources/{os.path.basename(temp_file)}"


def find_all_keys(d, target_key):
    if isinstance(d, dict):
        for key, value in d.items():
            if key == target_key:
                yield value
            elif isinstance(value, (dict, list, tuple)):
                yield from find_all_keys(value, target_key)
    elif isinstance(d, (list, tuple)):
        for item in d:
            yield from find_all_keys(item, target_key)


class OperatorModule:
    def __init__(self, sever: viser.ViserServer):
        self.upload_param_btn: Optional[viser.GuiUploadButtonHandle] = None
        self.upload_urdf_btn: Optional[viser.GuiUploadButtonHandle] = None
        self.server: viser.ViserServer = sever

        self.init_viser_module()

    def init_viser_module(self):
        self.upload_urdf_btn = self.server.gui.add_upload_button("上传urdf", color="blue")
        self.upload_param_btn = self.server.gui.add_upload_button("上传规划参数", color="blue")


def download_file(fname):
    try:
        """ 下载 HTTP 文件并返回本地路径 """
        response = requests.get(fname, stream=True)
        if response.status_code != 200:
            raise ValueError(f"无法下载文件: {fname}, 状态码: {response.status_code}")
        # 获取文件名（从 URL 提取）
        local_filename = os.path.basename(fname)

        # 创建临时目录存储文件
        temp_dir = tempfile.mkdtemp()
        file_path = os.path.join(temp_dir, local_filename)

        # 写入文件
        with open(file_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)

        return file_path

    except Exception as e:
        path = urllib.parse.urlparse(fname).path
        return path


class PlanModule:
    """
    规划场景模块，主要用来实现URDF的加载以及TF的计算和一些坐标系相关的内容可视化。
    """

    def __init__(self, server: viser.ViserServer, urdf_path):
        self.viser_urdf: Optional[ViserUrdf] = None
        self.server = server
        self.urdf: Optional[yourdfpy.URDF] = None
        self.frame: Optional[viser.FrameHandle] = None
        self.urdf_path = urdf_path

        self.init_viser_module()

    def init_viser_module(self):
        self.urdf = yourdfpy.URDF.load(self.urdf_path, filename_handler=download_file)

        self.frame = self.server.scene.add_frame(f"/world/{self.urdf.robot.name}", show_axes=False, visible=True)
        self.viser_urdf = ViserUrdf(self.server,
            urdf_or_path=self.urdf,
            root_node_name=f"/world/{self.urdf.robot.name}"
        )

        self.viser_urdf.update_cfg(np.array([0.0 for _ in self.urdf.actuated_joints]))

    def add_frame(self, frame_id: str, point):
        frames = self.viser_urdf._joint_frames
        parent_frame = [f"{frame.name}/frame" for frame in frames if frame.name.endswith(frame_id)][-1]

        self.server.scene.add_frame(parent_frame, position=np.array(point) * 0.001, axes_length=0.1, axes_radius=0.01)

    def add_box(self, frame_id: str, dimensions):
        frames = self.viser_urdf._joint_frames
        parent_frame = [f"{frame.name}" for frame in frames if frame.name.endswith(frame_id)][-1]

        box = np.array(dimensions) * 0.001
        position = box * 0.5
        position[1] *= -1
        self.server.scene.add_box(f"{parent_frame}/box", dimensions=box, position=position, color=(0, 255, 0))

    def set_ext(self, ext_value):
        joints = self.viser_urdf._urdf.actuated_joints
        value = ext_value * 0.001

        self.viser_urdf.update_cfg(np.array([value if joint.parent.endswith("zero") else 0.0 for joint in joints]))

    def get_transform(self, frame_to, frame_from):
        transform = self.viser_urdf._urdf.get_transform(frame_to, frame_from)

        return transform

    def remove(self):
        self.viser_urdf.remove()


class PlanParamModule:
    """
    规划参数模块，用来管理那些需要调节的参数，并且调用规划服务进行规划。
    """

    def __init__(self, server: viser.ViserServer, plan_param: dict):
        self.ext_range: list[viser.GuiVector3Handle] = []
        self.plan_param = plan_param
        self.plan_btn: Optional[viser.GuiButtonHandle] = None
        self.rotation: Optional[viser.GuiVector3Handle] = None
        self.server = server
        self.folder: Optional[viser.GuiFolderHandle] = None

        self.init_viser_module()

    def add_ext_joint_item(self, index: int):
        """
        添加外部轴返回调节模块，分别是上下界以及步长。
        :param index: 轴索引，0，1，2分别对应x轴，y轴，z轴。
        :return:
        """
        OPTIONS = ["X", "Y", "Z"]
        ext_joint_param = self.plan_param["#base"]["ext_joint_param"]["external_range"][index]
        ext_range = (self.server.gui.add_vector3(f"{OPTIONS[index]}范围",
            initial_value=np.array([ext_joint_param["lower"], ext_joint_param["step"], ext_joint_param["upper"]]),
            min=np.array([-10, 0, -10]),
            max=np.array([10, 10, 10]),
            step=0.2,
        ))
        ext_range.precision = 3

        @ext_range.on_update
        def _(event: viser.GuiEvent):
            """
            绑定更新事件，当数值更新的时候，更新规划参数里面的数据，方便规划的时候直接使用。
            :param event:
            :return:
            """
            value = event.target.value
            self.plan_param["#base"]["ext_joint_param"]["external_range"][index]["lower"] = value[0]
            self.plan_param["#base"]["ext_joint_param"]["external_range"][index]["step"] = value[1]
            self.plan_param["#base"]["ext_joint_param"]["external_range"][index]["upper"] = value[2]

        self.ext_range.append(ext_range)

    def init_viser_module(self):
        with self.server.gui.add_folder(self.plan_param["#base"]["id"][26:]) as self.folder:
            rotation_z = self.plan_param["#base"]["constraints"][0]["rotation"]["z_bound"]
            self.rotation = self.server.gui.add_vector3("自由度",
                initial_value=np.degrees(np.array([rotation_z["lower"], rotation_z["step"], rotation_z["upper"]])),
                min=np.degrees(np.array([-math.pi, 0, -math.pi])),
                max=np.degrees(np.array([math.pi, 2 * math.pi, math.pi])),
                step=5.0,
            )
            self.add_ext_joint_item(0)
            self.add_ext_joint_item(1)
            self.add_ext_joint_item(2)

            self.plan_btn = self.server.gui.add_button("重新规划")

            @self.rotation.on_update
            def _(_):
                value = np.radians(self.rotation.value)
                self.plan_param["#base"]["constraints"][0]["rotation"]["z_bound"]["lower"] = value[0]
                self.plan_param["#base"]["constraints"][0]["rotation"]["z_bound"]["step"] = value[1]
                self.plan_param["#base"]["constraints"][0]["rotation"]["z_bound"]["upper"] = value[2]

            self.folder.expand_by_default = False

    def remove(self):
        self.server.scene.remove_by_name(self.folder.name)
        self.server.gui.reset()


class SimulationModule:
    """
    仿真模块，主要包括用来查看仿真的几个控件，比如滑块，一些按钮等。
    """

    def __init__(self, server: viser.ViserServer = None, sim_robot: ViserUrdf = None):
        self.timestep: Optional[viser.GuiSliderHandle[int]] = None
        self.play: Optional[viser.GuiButtonHandle] = None
        self.prev_frame: Optional[viser.GuiButtonHandle] = None
        self.next_frame: Optional[viser.GuiButtonHandle] = None
        self.trajectory: Optional[dict] = None
        self.sim_robot: ViserUrdf = sim_robot

        self.tool: Optional[viser.MeshHandle] = None
        self.workpiece: Optional[viser.MeshHandle] = None

        self.server = server

    @property
    def server(self) -> viser.ViserServer:
        return self._server

    @server.setter
    def server(self, server: viser.ViserServer):
        if server is not None:
            self._server = server
            self.init_viser_module()

    def init_viser_module(self):
        with self.server.gui.add_folder("Simulation") as self.folder:
            self.timestep = self.server.gui.add_slider(
                "Timestep",
                min=0,
                max=1,
                step=1,
                initial_value=0,
                disabled=False,
            )

            self.next_frame = self.server.gui.add_button("Next Frame")
            self.prev_frame = self.server.gui.add_button("Prev Frame")
            self.play = self.server.gui.add_button("Play")

            @self.next_frame.on_click
            def _(_):
                self.timestep.value = min(self.timestep.value + 1, int(self.timestep.max))

            @self.prev_frame.on_click
            def _(_):
                self.timestep.value = max(self.timestep.value - 1, 0)

            @self.play.on_click
            def _(_):
                self.replay_trajectory()

            @self.timestep.on_update
            def _(_):
                self.update_robot(int(self.timestep.value))

            self.folder.expand_by_default = False

    def update_robot(self, index):
        # 根据轨迹做映射更新机器人状态
        joint_names: list[str] = self.trajectory["joint_names"]
        trajectory: list[float] = self.trajectory["trajectory"][index]["position"]
        self.sim_robot.update_cfg(
            np.array([trajectory[joint_names.index(joint.name)] for joint in self.sim_robot._urdf.actuated_joints]))

    def replay_trajectory(self):
        for i in range(len(self.trajectory["trajectory"])):
            self.update_robot(i)
            time.sleep(0.1)

    def set_trajectory(self, trajectory: dict):
        self.trajectory = trajectory

        self.timestep.max = len(self.trajectory["trajectory"]) - 1

    def load_param(self, plan_param):
        """
        通过规划参数加载起始状态，以及工件工具头。
        :param plan_param:
        :return:
        """
        if self.tool is not None:
            self.tool.remove()
        if self.workpiece is not None:
            self.workpiece.remove()
        if self.sim_robot is None:
            return
        frames = self.sim_robot._joint_frames
        tool = plan_param["#base"]["tool_mesh"]
        tool_shape = tool["shape"]
        workpiece = plan_param["#base"]["workpiece_mesh"]
        wp_mesh = workpiece["shape"]
        tool_frame = [f"{frame.name}/{tool['name']}" for frame in frames if frame.name.endswith(tool['frame_id'])][-1]
        wp_frame = \
            [f"{frame.name}/{workpiece['name']}" for frame in frames if frame.name.endswith(workpiece['frame_id'])][-1]
        self.tool = self.server.scene.add_mesh_simple(tool_frame, np.array(tool_shape["vertices_m"]).T,
            np.array(tool_shape["face_m"]).T)
        self.workpiece = self.server.scene.add_mesh_simple(wp_frame, np.array(wp_mesh["vertices_m"]).T,
            np.array(wp_mesh["face_m"]).T)


class Punch:
    def __init__(self):
        self.server = viser.ViserServer(label="Punch一拳")

        self.operator = OperatorModule(self.server)

        self.simulation = SimulationModule()
        self.plan_module: Optional[PlanModule] = None
        self.plan_param_module: Optional[PlanParamModule] = None

        self.init_viser()

    def init_viser(self):
        self.server.scene.add_grid("/world")

        self.server.gui.configure_theme(control_width="large")

        @self.operator.upload_urdf_btn.on_upload
        def _(event):
            """上传URDF的时候把规划模块初始化好，主要是加载URDF"""
            file = self.operator.upload_urdf_btn.value

            file_path = make_resource_dir(".urdf")

            open(file_path, "wb").write(file.content)

            self.plan_module = PlanModule(event.client, file_path)

            self.simulation.sim_robot = self.plan_module.viser_urdf
            self.simulation.server = self.plan_module.server

        @self.operator.upload_param_btn.on_upload
        def _(event):
            """当规划参数文件上传的时候，读取规划参数，并且把可调参数通过UI展示."""
            file = self.operator.upload_param_btn.value

            file_path = make_resource_dir(".json")

            self.plan_param_module = PlanParamModule(event.client, json.loads(file.content))

            self.simulation.load_param(self.plan_param_module.plan_param)

            @self.plan_param_module.plan_btn.on_click
            def _(_):
                # 点击重规划时，把当前使用的URDF以及更新过的规划参数直接调用规划服务。
                json.dump(self.plan_param_module.plan_param, open(file_path, "w"))

                res = requests.get(
                    f"http://127.0.0.1:12123/plan?urdf={self.plan_module.urdf_path}&param={file_path}")

                self.simulation.set_trajectory(res.json())

    def run(self):
        while True:
            time.sleep(0.1)


def main() -> None:
    process = subprocess.Popen(
        ["uvicorn", "run:app", "--host", "127.0.0.1", "--port", "8000", "--reload"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    punch = Punch()

    punch.run()


if __name__ == "__main__":
    main()
