from datetime import datetime
import io
import json
import os
import shutil
import time
import zipfile
from typing import Optional

import numpy as np
import urdf_parser_py.urdf
import viser

from punch.plan_client import PlanModule
from robim_data import Workshop
from .urdf import URDFRobot


def unzip_pes(pes_file: str, extract_path: str):
    os.makedirs(extract_path, exist_ok=True)
    with zipfile.ZipFile(pes_file, 'r') as zip_ref:
        zip_ref.extractall(extract_path)


class PesScene:
    """
    PES场景模块，主要复杂加载PES文件里面的scene.json和workshop.proto。并且在后续更新掉URDF之后保存成新的PES文件。
    """

    def __init__(self, resource_dir: str):
        self.workshop = Workshop()
        self.resource_dir = resource_dir
        self.scene = json.load(open(f"{resource_dir}/scene.json"))

        self.name = self.scene["name"]

        # 读取二进制文件
        with open(f"{resource_dir}/workshop.proto", "rb") as f:
            data = f.read()

        self.workshop.ParseFromString(data)
        self._urdf = None

        self.urdf = self.workshop.urdf

        self.urdf_robot: URDFRobot = URDFRobot.from_xml_string(self.urdf)
        # 对应型钢这样的标准单臂场景，base表示机器人坐标系，也就是戳点对应的坐标系，attached表示小车坐标系，也就是工件放的坐标系。
        self.attached = [link.name for link in self.urdf_robot.links if link.name.endswith("attached")][-1]
        self.robot_base = [link.name for link in self.urdf_robot.links if link.name.endswith("base")][-1]

        self.update_name()

        self.urdf_robot.name = self.name
        self.scene["name"] = self.name

    def update_name(self):
        """
        更新场景的名称，如果场景的名称包含下划线，说明之前已经添加过日期了，则把日期修改成最新日期。否则直接添加日期。
        :return:
        """
        self.name = self.scene["name"]
        # 找到最后一个下划线的位置
        pos = self.name.rfind("_")

        date_str = datetime.now().strftime("%y%m%d")
        # 如果找到下划线，则进行替换
        if pos != -1:
            self.name = self.name[:pos + 1] + date_str
        else:
            self.name = self.name + "_" + date_str  # 没有下划线则不做修改

    @property
    def urdf(self):
        return self._urdf

    @urdf.setter
    def urdf(self, value):
        self._urdf = value

        self.workshop.urdf = value
        self.scene['rpcpUrdf'] = value

    def save(self):
        open(f"{self.resource_dir}/scene.json", "w").write(json.dumps(self.scene))
        open(f"{self.resource_dir}/workshop.proto", "wb").write(self.workshop.SerializeToString())

    def zip(self):
        pes_file = f"{os.path.dirname(self.resource_dir)}/{self.name}.pes"
        with zipfile.ZipFile(pes_file, "w", zipfile.ZIP_DEFLATED,
                allowZip64=True) as zip_ref:
            for root, dirs, files in os.walk(self.resource_dir):
                for file in files:
                    file_path = os.path.join(root, file)
                    arcname = os.path.relpath(file_path, self.resource_dir)  # 相对路径
                    zip_ref.write(file_path, arcname)

        return pes_file


class PointDataModule:
    def __init__(self, server: viser.ViserServer):
        self.server = server
        self.folder: Optional[viser.GuiFolderHandle] = None

        self.point_data: list[tuple[viser.GuiVector3Handle, viser.GuiNumberHandle]] = []

        self.init_viser_module()

    def init_viser_module(self):
        with self.server.gui.add_folder("戳点数据") as self.folder:
            self.folder.expand_by_default = True

        self.add_point()

    def add_point(self):
        with self.folder:
            self.point_data.append((
                self.server.gui.add_vector3(f"机器人戳点{len(self.point_data)}",
                    initial_value=np.array([829.819, -561.124, 612.729]),
                    step=100),
                self.server.gui.add_number(f"外部轴值{len(self.point_data)}", initial_value=-525.008, step=100.0),))

            self.point_data[-1][0].precision = 3
            self.point_data[-1][1].precision = 3

    def remove_point(self):
        if len(self.point_data) == 0:
            return
        point_data = self.point_data[-1]
        point_data[0].visible = False
        point_data[1].visible = False

        self.point_data.pop()


class CalibrationResultModule:
    def __init__(self, server: viser.ViserServer):
        self.server = server
        # 计算按钮
        self.calc_btn: Optional[viser.GuiButtonHandle] = None
        # 计算出来的戳点在小车坐标系下的点位，如果准的话，应该是工件尺寸一样的。
        self.xyz: Optional[viser.GuiVector3Handle] = None
        # 计算出来的更新偏差。后续直接作用更新。
        self.diff: Optional[viser.GuiVector3Handle] = None
        # 分别是3个按钮，用来分别更新xyz数值。因为场景大部分是横平竖直的。
        self.apply_btns: Optional[viser.GuiButtonGroupHandle] = None
        # 下载按钮，用于下载更新过后的场景，之后应用过更改才提供下载功能。
        self.download: Optional[viser.GuiButtonHandle] = None

        self.init_viser_module()

    def init_viser_module(self):
        with self.server.gui.add_folder("标定结果") as self.folder:
            self.calc_btn = self.server.gui.add_button("计算")
            self.xyz = self.server.gui.add_vector3("计算结果", initial_value=np.array([0, 0, 0]))
            self.xyz.precision = 3
            self.diff = self.server.gui.add_vector3("偏差", initial_value=np.array([0, 0, 0]))
            self.diff.precision = 3
            self.apply_btns = self.server.gui.add_button_group("应用", options=["X", "Y", "Z"])
            self.download = self.server.gui.add_button("下载", disabled=True)


class OperatorModule:
    def __init__(self, sever: viser.ViserServer):
        self.add_remove: Optional[viser.GuiButtonGroupHandle] = None
        self.workpiece_info: Optional[viser.GuiVector3Handle] = None
        self.update_pes_btn: Optional[viser.GuiUploadButtonHandle] = None
        self.server: viser.ViserServer = sever

        self.init_viser_module()

    def init_viser_module(self):
        self.update_pes_btn = self.server.gui.add_upload_button("上传场景", color="blue")

        self.workpiece_info = self.server.gui.add_vector3("工件尺寸", initial_value=np.array([1000, 200, 50]),
            min=np.array([0, 0, 0]), max=np.array([1e10, 1e10, 1e10]), step=100, hint="工件尺寸")
        self.workpiece_info.precision = 3

        self.add_remove = self.server.gui.add_button_group("操作", ["+", "-"])


class PlanSceneUpdater:
    def __init__(self):
        self.server = viser.ViserServer(label="场景更新工具")

        self.operator_module = OperatorModule(self.server)

        self.scene_modules: dict[viser.ClientHandle, PesScene] = {}

        self.plan_modules: dict[viser.ClientHandle, PlanModule] = {}

        self.point_data_modules: dict[viser.ClientHandle, PointDataModule] = {}
        self.result_modules: dict[viser.ClientHandle, CalibrationResultModule] = {}

        self.initialize()

        @self.server.on_client_connect
        def _(client: viser.ClientHandle):
            self.point_data_modules[client] = PointDataModule(client)
            self.result_modules[client] = CalibrationResultModule(client)

            @self.result_modules[client].calc_btn.on_click
            def _(event: viser.GuiEvent):
                """
                计算偏差，主要原理是通过把戳的点通过坐标系变换变换到小车坐标下，然后跟工件尺寸进行对比。最后转换回世界坐标系准备更新。
                :param event:
                :return:
                """
                self.process_scene(client=client, )
                scene = self.scene_modules[client]
                transform = self.plan_modules[client].get_transform(scene.robot_base, scene.attached)

                point = self.point_data_modules[client].point_data[-1][0].value
                workpiece_dimensions = np.array(self.operator_module.workpiece_info.value)

                calc_point = transform[:3, :3] @ point + transform[:3, 3] * 1000

                # 获取世界坐标系的变换
                transform = self.plan_modules[client].get_transform(scene.attached, "base_link")

                workpiece_dimensions[1] *= -1
                self.result_modules[client].xyz.value = calc_point
                diff = workpiece_dimensions - calc_point
                diff = transform[:3, :3] @ diff
                self.result_modules[client].diff.value = diff

            @self.result_modules[client].apply_btns.on_click
            def _(event: viser.GuiEvent):
                scene = self.scene_modules[client]
                parent_map = scene.urdf_robot.parent_map

                need_to_update = parent_map[parent_map[parent_map[scene.attached][-1]][-1]][0]

                need_to_update = scene.urdf_robot.joint_map[need_to_update]
                btn = self.result_modules[client].apply_btns
                index = btn.options.index(btn.value)
                need_to_update.origin.position[index] -= self.result_modules[client].diff.value[index] * 0.001

                scene.urdf = scene.urdf_robot.to_xml_string()

                self.plan_modules[event.client].remove()
                self.plan_modules[event.client] = PlanModule(event.client,
                    io.BytesIO(self.scene_modules[event.client].urdf.encode()))
                self.process_scene(client=client)
                self.result_modules[client].download.disabled = False

            @self.result_modules[client].download.on_click
            def _(event: viser.GuiEvent):
                self.scene_modules[client].save()
                file = self.scene_modules[client].zip()
                client.send_file_download(f"{self.scene_modules[client].name}.pes", open(file, "rb").read())

    def process_scene(self, client: viser.ClientHandle):
        """
        通过输入的工件尺寸以及点位，在场景中加入可视化控件。包括在推料小车上增加一个工件尺寸大小的盒子，并且把戳的点画出来。同时更新掉推料小车的位置。
        :return:
        """
        point = self.point_data_modules[client].point_data[-1][0].value
        workpiece_dimensions = self.operator_module.workpiece_info.value
        self.plan_modules[client].add_box(self.scene_modules[client].attached, workpiece_dimensions)
        self.plan_modules[client].add_frame(self.scene_modules[client].robot_base, point)
        self.plan_modules[client].set_ext(self.point_data_modules[client].point_data[-1][1].value)

    def initialize(self):
        self.server.scene.add_grid("/world")

        self.server.gui.configure_theme(control_width="large")

        @self.operator_module.update_pes_btn.on_upload
        def _(event: viser.GuiEvent):
            """Callback for when a file is uploaded."""
            file = self.operator_module.update_pes_btn.value

            open(f"resources/{file.name}", "wb").write(file.content)

            resource_dir = f"resources/{os.path.splitext(file.name)[0]}"

            shutil.rmtree(resource_dir, ignore_errors=True)

            unzip_pes(f"resources/{file.name}", resource_dir)

            self.scene_modules[event.client] = PesScene(resource_dir)

            self.plan_modules[event.client] = PlanModule(event.client,
                io.BytesIO(self.scene_modules[event.client].urdf.encode()))

        @self.operator_module.add_remove.on_click
        def _(event: viser.GuiEvent):
            if self.operator_module.add_remove.value == '+':
                self.point_data_modules[event.client].add_point()
            else:
                self.point_data_modules[event.client].remove_point()

    @staticmethod
    def run():
        while True:
            time.sleep(0.1)


def main():
    updater = PlanSceneUpdater()
    updater.run()
